

################################################################################


class Magnetization:
  

#------------------------------------------------------------------------------#


  def __init__(self, geometry, V, X, m0, maxwellCoupling, solvetol, j_cf):
    import sys
    from ngsolve import GridFunction
    from . import strayField, oerstedField, _zeroField

    # X is the compound FESpace \subset H1(\Omega)^3
    self._mesh = geometry.GetMesh()
    
    self.maxwellCoupling = maxwellCoupling
    
    self.m = GridFunction(X)
    self.Dm = self.m.Deriv()
    self.cf = self.m
   
    if( self.maxwellCoupling == "magnetostaticField" ):
      self.hs = strayField.StrayField(self._mesh, self.cf, solvetol)
      if j_cf != None:
        self.hc = oerstedField.OerstedField(self._mesh, j_cf, solvetol)
      else:
        self.hc = _zeroField.ZeroField()
    elif( self.maxwellCoupling == "strayField" ):
      self.hs = strayField.StrayField(self._mesh, self.cf, solvetol)
    elif( self.maxwellCoupling == "oerstedField" ):
      if j_cf != None:
        self.hc = oerstedField.OerstedField(self._mesh, j_cf, solvetol)
      else:
        self.hc = _zeroField.ZeroField()
    elif( self.maxwellCoupling == "eddyCurrent" ):
      pass
    elif( self.maxwellCoupling != "none" ): 
      print("Error in ._Mag.__init__():")
      print("Maxwell options not >strayField<, >oerstedField<, >eddyCurrent< \
        or >none< are not implemented yet")
      sys.exit()
   
    self.SetInitialValue(m0)


#------------------------------------------------------------------------------#


  def SetInitialValue(self, m0):

    if isinstance(m0, str):
      self.SetFromFile(m0)
    elif isinstance(m0, type(self)):
      self.SetFromOther(m0)
    else:
      self.SetFromCF(m0)


#------------------------------------------------------------------------------#


  def SetFromCF(self, nu):
    from ngsolve import CoefficientFunction
    for i in range(3):
      if (isinstance(nu[i], int) or isinstance(nu[i], float)):
        self.m.components[i].Set( CoefficientFunction(nu[i]) )
      else:
        self.m.components[i].Set( nu[i] )
  
  
#------------------------------------------------------------------------------#


  def SetFromFile(self, filename):
    from commix._tools import LoadMagnetization
    self.m.vec.FV().NumPy()[:] = LoadMagnetization( filename )
  
  
#------------------------------------------------------------------------------#


  def SetFromOther(self, other, withTimeDerivative=False):
    import sys
    self.m.vec.FV().NumPy()[:] = other.m.vec.FV().NumPy()
    
    if ( self.maxwellCoupling != other.maxwellCoupling ):
      print("Error in ._Mag.SetFromOther():")
      print("Maxwell options of other and self don't match!")
      sys.exit()
    
    if ( self.maxwellCoupling == "magnetostaticField" ):
      self.hs.SetFromOther(other.hs)
      self.hc.SetFromOther(other.hc)
    elif ( self.maxwellCoupling == "strayField" ):
      self.hs.SetFromOther(other.hs)
    elif ( self.maxwellCoupling == "oerstedField" ):
      self.hc.SetFromOther(other.hc)
    elif ( self.maxwellCoupling == "eddyCurrent" ):
      self.H.SetFromOther(other.H)
    elif ( self.maxwellCoupling != "none" ):
      print("Error in ._Mag.SetFromOther():")
      print("Maxwell options not >strayField<, >oerstedField<, >eddyCurrent< \
        or >none< are not implemented yet")
      sys.exit()
    
    if (withTimeDerivative):
      self.v.gf.vec.FV().NumPy()[:] = other.v.gf.vec.FV().NumPy()
  
  
#------------------------------------------------------------------------------#


  def UpdateMaxwellCoupling(self, newMaxwellCoupling, solvetol, j_cf):
    from . import strayField, oerstedField, _zeroField

    self.maxwellCoupling = newMaxwellCoupling
    self.hs = None
    self.hc = None

    if( self.maxwellCoupling == "magnetostaticField" ):
      self.hs = strayField.StrayField(self._mesh, self.cf, solvetol)
      if j_cf != None:
        self.hc = oerstedField.OerstedField(self._mesh, j_cf, solvetol)
      else:
        self.hc = _zeroField.ZeroField()
    elif( self.maxwellCoupling == "strayField" ):
      self.hs = strayField.StrayField(self._mesh, self.cf, solvetol)
    elif( self.maxwellCoupling == "oerstedField" ):
      if j_cf != None:
        self.hc = oerstedField.OerstedField(self._mesh, j_cf, solvetol)
      else:
        self.hc = _zeroField.ZeroField()
    elif( self.maxwellCoupling == "eddyCurrent" ):
      pass
    elif( self.maxwellCoupling != "none" ): 
      print("Error in ._Mag.__init__():")
      print("Maxwell options not >strayField<, >oerstedField<, >eddyCurrent< \
        or >none< are not implemented yet")
      sys.exit()

    
#------------------------------------------------------------------------------#


  def Project(self):
    from scipy import sqrt
    # projects m onto Sphere pointwise

    mx = self.m.components[0].vec.FV().NumPy()
    my = self.m.components[1].vec.FV().NumPy()
    mz = self.m.components[2].vec.FV().NumPy()
    scale = sqrt(mx*mx + my*my + mz*mz)
    self.m.components[0].vec.FV().NumPy()[:] = mx / scale
    self.m.components[1].vec.FV().NumPy()[:] = my / scale
    self.m.components[2].vec.FV().NumPy()[:] = mz / scale
    
  
#------------------------------------------------------------------------------#


  def ComputeMaxwell(self, solvetol):
    import sys

    if(self.maxwellCoupling == "magnetostaticField" ):
      self._ComputeStrayField(solvetol)
      self._ComputeOerstedField(solvetol)
    elif(self.maxwellCoupling == "strayField"):
      self._ComputeStrayField(solvetol)
    elif(self.maxwellCoupling == "oerstedField"):
      self._ComputeOerstedField(solvetol)
    elif(self.maxwellCoupling != "none"):
      print("Error in _Mag.ComputeMaxwell:")
      print("maxwellCoupling =", self.maxwellCoupling, "unknown.")
      sys.exit()
  

#------------------------------------------------------------------------------#


  def _ComputeMagnetostaticField(self, solvetol):
    import sys
    if (self.maxwellCoupling != "magnetostaticField" ):
      print("Error in _Mag._ComputeMagnetostaticField():")
      print("maxwellCoupling != <magnetostaticField> !! --> Terminate.")
      sys.exit()
        
    self._ComputeStrayField(solvetol)
    self._ComputeOerstedField(solvetol)
  

#------------------------------------------------------------------------------#


  def _ComputeStrayField(self, solvetol):
    import sys
    if ( self.maxwellCoupling != "strayField" \
      and self.maxwellCoupling != "magnetostaticField" ):
      print("Error in _Mag._ComputeStrayField():")
      print("maxwellCoupling != <strayField> !! --> Terminate.")
      sys.exit()
    
    self.hs.Compute(solvetol)
  

#------------------------------------------------------------------------------#


  def _ComputeOerstedField(self, solvetol):
    import sys
    if ( self.maxwellCoupling != "oerstedField" \
      and self.maxwellCoupling != "magnetostaticField" ):
      print("Error in _Mag._ComputeOerstedField():")
      print("maxwellCoupling != <oerstedField> !! --> Terminate.")
      sys.exit()
    
    self.hc.Compute(solvetol)
  

#------------------------------------------------------------------------------#


  def InitializeEddyCurrent(self):
    import sys
    if ( self.maxwellCoupling != "eddyCurrent" ):
      print("Error in _Mag.ComputeEddyCurrent():")
      print("maxwellCoupling != <eddyCurrent> !! --> Terminate.")
      sys.exit()
    
    self.H.SetFromMagnetization(self.m)
  

#------------------------------------------------------------------------------#


  def phiUniaxialAnisotropy(self, e):
    from commix._tools.algebra import InnerProd
    return -1.0 * InnerProd(e, self.m) * e
  
  
#------------------------------------------------------------------------------#


  def opZhangLi(self, vel):
    return (sum(vel[k]*self.Dm[0,k] for k in range(3)), \
      sum(vel[k]*self.Dm[1,k] for k in range(3)), \
      sum(vel[k]*self.Dm[2,k] for k in range(3)) )
  
  
#------------------------------------------------------------------------------#


  def Cross(self, u):
    from commix._tools.algebra import CrossProd
    return CrossProd(self.m, u)
  
  
#------------------------------------------------------------------------------#


  def Draw(self, nameString = "m"):
    from ngsolve import Draw
    Draw(self.cf, self._mesh, nameString)
  
  
#------------------------------------------------------------------------------#


  def Averages(self):
    from ngsolve import Integrate, CoefficientFunction
    measOmega = Integrate(CoefficientFunction(1.0), self._mesh)
    
    ngAvg = Integrate(self.m, self._mesh)

    # cast from ngs-type to list and scale
    avg = [2.0, 2.0, 2.0]
    for i in range(3): avg[i] = ngAvg[i] / measOmega
    
    return avg
  
  
#------------------------------------------------------------------------------#


  def Norms(self):
    nrm = [-1.0, -1.0]
    nrm[0] = self.NormLinfty()
    nrm[1] = self.NormL1mSquaredminus1()
    
    return nrm
  
  
#------------------------------------------------------------------------------#


  def NormLinfty(self):
    from scipy import sqrt
    # compute L^infty(Omega)-Norm
    
    mx = self.m.components[0].vec.FV().NumPy()
    my = self.m.components[1].vec.FV().NumPy()
    mz = self.m.components[2].vec.FV().NumPy()
    modulus = sqrt(mx*mx + my*my + mz*mz)

    return modulus.max()
  
  
#------------------------------------------------------------------------------#


  def NormL1mSquaredminus1(self):
    from ngsolve import IfPos, GridFunction, Integrate
    # compute L^1(Omega)-Norm of |m|^2-1
    
    tmp_gf = GridFunction(self.m.space.components[0]) 
    tmp_gf.vec.FV().NumPy()[:] = 0.0
    
    for i in range(3):
      tmp_gf.vec.FV().NumPy()[:] += self.m.components[i].vec.FV().NumPy() \
        * self.m.components[i].vec.FV().NumPy()
    
    fct = IfPos( tmp_gf-1.0, tmp_gf-1.0, -(tmp_gf-1.0) )
    nrm = Integrate(fct, self._mesh)
    
    return nrm
  
  
#------------------------------------------------------------------------------#


  def PointwiseAverages(self):
    N = self.m.components[0].vec.size
    
    avg = [2.0, 2.0, 2.0]
    for i in range(3): avg[i] = self.m.components[i].vec.FV().NumPy().sum() / N 
    
    return avg
  
  
#------------------------------------------------------------------------------#


  def PrintAverages(self, t):
    from commix._tools.misc import TimeToString
    avg = self.Averages()
    print(" <m> = [%.3f" % avg[0], " %.3f" % avg[1], " %.3f" % avg[2] \
      , "] at " + TimeToString(t), sep='')
  
  
#------------------------------------------------------------------------------#


  def EnergyExchange(self):
    from commix._tools.algebra import FrobeniusInnerProd
    from ngsolve import Integrate
    energy = Integrate(FrobeniusInnerProd(self.Dm, self.Dm), self._mesh)
    return energy
  
  
#------------------------------------------------------------------------------#


  def EnergyUniaxialAnisotropy(self, easy_cf):
    from ngsolve import CoefficientFunction, Integrate
    # TODO catch if easy_cf = None
    energy = Integrate( CoefficientFunction(1.0) \
    - (easy_cf*self.cf)*(easy_cf*self.cf), self._mesh)
    return energy
  
  
#------------------------------------------------------------------------------#


  def EnergyBulkDzyaloshinskiiMoriya(self):
    from ngsolve import Integrate
    energy = Integrate( \
      ( self.Dm[2,1] \
      - self.Dm[1,2] ) * self.cf[0] \
      + ( self.Dm[0,2] \
      - self.Dm[2,0] ) * self.cf[1] \
      + ( self.Dm[1,0] \
      - self.Dm[0,1] ) * self.cf[2], \
      self._mesh )
    return energy
  
  
#------------------------------------------------------------------------------#


  def EnergyInterfacialDzyaloshinskiiMoriya(self):
    from ngsolve import Integrate
    energy = Integrate( \
      (self.Dm[0,0] + self.Dm[1,1]) \
      * self.cf[2] \
      - self.Dm[2,0] * self.cf[0] \
      - self.Dm[2,1] * self.cf[1], \
      self._mesh )
    return energy
  
  
#------------------------------------------------------------------------------#


  def EnergyStrayField(self):
    import sys
    from ngsolve import Integrate
  # TOECDO zu energy maxwell zusammenlegen
    if ( self.maxwellCoupling != "strayField" \
      and self.maxwellCoupling != "magnetostaticField" ):
      print("Error in _Mag.EnergyStrayField():")
      print("maxwellCoupling != <strayField> !! --> Terminate.")
      sys.exit()
    
    energy = Integrate( self.hs.cf * self.cf, self._mesh)
    return energy
  
  
#------------------------------------------------------------------------------#


  def EnergyOerstedField(self):
    import sys
    from ngsolve import Integrate
  # TOECDO zu energy maxwell zusammenlegen
    if ( self.maxwellCoupling != "oerstedField" \
      and self.maxwellCoupling != "magnetostaticField" ):
      print("Error in _Mag.EnergyOerstedField():")
      print("maxwellCoupling != <oerstedField> !! --> Terminate.")
      sys.exit()
    
    energy = Integrate( self.hc.cf * self.cf, self._mesh)
    return energy
  
  
#------------------------------------------------------------------------------#


  def EnergyEddyCurrent(self):
    import sys
    from ngsolve import Integrate
    if ( self.maxwellCoupling != "eddyCurrent" ):
      print("Error in _Mag.EnergyEddyCurrent():")
      print("maxwellCoupling != <eddyCurrent> !! --> Terminate.")
      sys.exit()
    
    energy = Integrate( self.H.gf * self.H.gf, self.H.GetMesh())
    return energy
  
  
#------------------------------------------------------------------------------#


  def EnergyApplied(self, hext_cf):
    from ngsolve import Integrate
    if hext_cf == None:
      return 0.0

    energy = Integrate( hext_cf * self.cf, self._mesh)
    return energy
  
  
#------------------------------------------------------------------------------#


  def Save(self, filepath):
    from commix._tools import SaveMagnetization
    SaveMagnetization(filepath, self._mesh, self.m)
#    if ( self.maxwellCoupling == "eddyCurrent" ): # TOECDO
#      self.H.Save(filepath)
  
  
#------------------------------------------------------------------------------#


  def SaveStrayField(self, filepath):
    import sys
    if(self.maxwellCoupling != "strayField" and \
      self.maxwellCoupling != "magnetostaticField"):
      print("Error in _Mag.SaveStrayField() , wrong maxwellCoupling")
      print("-->Terminate")
      sys.exit()
    
    self.hs.Save(filepath)
  
  
#------------------------------------------------------------------------------#


  def SaveOerstedField(self, filepath):
    import sys
    if(self.maxwellCoupling != "oerstedField" and \
      self.maxwellCoupling != "magnetostaticField"):
      print("Error in _Mag.SaveOerstedField() , wrong maxwellCoupling")
      print("-->Terminate")
      sys.exit()
    
    self.hc.Save(filepath)
  
  
#------------------------------------------------------------------------------#


################################################################################
