from . import _details

################################################################################


class _Integrator:
  

#------------------------------------------------------------------------------#


  def __init__(self, geometry, parameters = None, output = None, numthreads = 1):

    # must be called before all scipy, numpy, ngsolve, bempp, ... imports
    self._SetNumThreads(numthreads)
    self._numthreads = numthreads

    from ngsolve.ngstd import Timer
    from ngsolve import ngsglobals
    from .. import _setup
    ngsglobals.msg_level = 0

    self._geometry = geometry
    self.parameters = (_setup.Parameters() if (parameters == None) else parameters)
    print(parameters)
    self.output = (_setup.OutputOptions() if (output == None) else output)
    
    self._nonTrivialMembers = ("_geometry", "parameters", "output")
    
    self._Mag = None
    
    self._C_ex = None
    self._C_ani = None
    self._C_st = None
    self._C_dm = None
    self._k = None
    self._T_end = None
    self._t = None
    self._time_cf = None
    self._hext_cf = None
    self._hext_impl_cf = None
    self._easy_cf = None
    self._p_cf = None
    self._j_cf = None
    self._jprime_cf = None
    
    self._V = None
    self._X = None

    self._psi = None
    self._phi = None
    
    self._Dpsi = None
    self._Dphi = None

    self._a_stat = None
    self._A_stat = None
    self._a_instat = None
    self._A_instat = None

    self._callback = None
    self._itCounter = _details.IterationCounter()
    self._preconditioner = None
    self._guess = None
    self._solution = None
    
    self._f_explicit = None
    
    self._LHS = None
    self._RHS = None

    self._newFolder = True
    
    self._compTimeNow = 0.0
    self._compTimeTotal = 0.0
        
    self._counter = 0 
    self._startCounter = 0
    self._steps = 0 

    self._maxwellTimer = Timer("magnetostaticField")

    self._saveMagEvery = None

    self._maxwellCouplingChanged = False


#------------------------------------------------------------------------------#


  def __str__(self):
    string = "\n\n -------------------- "
    string += "Integrator: " + self.__class__.__name__ 
    string += " --------------------" + "\n\n"
    string += "numthreads = " + str(self._numthreads) + "\n\n"
    
    for key, value in self.__dict__.items():
      if not(key.startswith("_") or key in self._nonTrivialMembers):
        string += key + " = " + str(value) + "\n"
    
    for key in self._nonTrivialMembers:
      string += self.__getattribute__(key).__str__() + "\n"

    string += "\n\n -------------------"
    string += "--------------------"
    string += "--------------------"
    string += "--------------------" + "\n\n"
    return string


#------------------------------------------------------------------------------#


  def _GetMesh(self):
    return self._geometry.GetMesh()
  

#------------------------------------------------------------------------------#
  

  def GetReady(self):

    self._CheckForBadParameterCombinations()

    self._PhysicalToMathematical()

    if(self._counter == 0):
      self._geometry.GetReady()
      self._SetUpSpaces()
      self._InitializeMagnetization()
      self._SetUpAuxiliaryFunctions()
      self._CreateOutputFolder()
    elif(self._maxwellCouplingChanged):
      self._UpdateMaxwellCoupling()
    self._maxwellCouplingChanged = False

    self._SetUpOutput()

    self._SetUpSystems()
    self._SetUpSchemeSpecificObjects()
    self._SetUpPreconditioner()


#------------------------------------------------------------------------------#


  def _UpdateMaxwellCoupling(self):

    if self.parameters.spintronicsCoupling == "slonczewski":
      self._Mag.UpdateMaxwellCoupling(self.parameters.maxwellCoupling \
        , self._solvetol, self._j_cf * self._p_cf)
    else:
      self._Mag.UpdateMaxwellCoupling(self.parameters.maxwellCoupling \
        , self._solvetol, self._j_cf)


#------------------------------------------------------------------------------#


  def _SetUpSchemeSpecificObjects(self): pass


#------------------------------------------------------------------------------#


  def _CheckForBadParameterCombinations(self):
    import sys
    ## TODO search for all the prints and sys.exits --> replace with raise(...)

    if(self.parameters.maxwellCoupling == "oerstedField" \
      or self.parameters.maxwellCoupling == "magnetostaticField"):
      print("maxwellCoupling set to", self.parameters.maxwellCoupling)
      print("Oersted field implementation is Work In Progress; not finished.")
      sys.exit()

    if(False): # TODO check bad combinations
      print("message TODO")
      sys.exit()


#------------------------------------------------------------------------------#


  _PhysicalToMathematical = _details.PhysicalToMathematical


#------------------------------------------------------------------------------#


  def _SetUpSpaces(self):
    from ngsolve import H1, VectorH1, ORDER_POLICY
    self._V = H1(self._geometry.GetMesh(), order=1)
    self._X = VectorH1(self._geometry.GetMesh(), order=1 \
      , order_policy=ORDER_POLICY.CONSTANT)

    self._psi = self._X.TrialFunction()
    self._phi = self._X.TestFunction()

    self._Dpsi = self._psi.Deriv()
    self._Dphi = self._phi.Deriv()



#------------------------------------------------------------------------------#


  def _InitializeMagnetization(self):
    from commix._gridFunctions import Magnetization
    if self.parameters.spintronicsCoupling == "slonczewski":
      self._Mag = Magnetization(self._geometry, self._V, self._X \
        , self.parameters.m0 \
        , self.parameters.maxwellCoupling, self._solvetol, self._j_cf * self._p_cf)
    else:
      self._Mag = Magnetization(self._geometry, self._V, self._X \
        , self.parameters.m0 \
        , self.parameters.maxwellCoupling, self._solvetol, self._j_cf )



#------------------------------------------------------------------------------#


  def _SetUpAuxiliaryFunctions(self):
    import sys
    print("Calling _SetUpAuxiliaryFunctions from _integrator")
    print("I don't know what to do, I'm only the base class --> Terminate")
    sys.exit()


#------------------------------------------------------------------------------#


  def _SetUpSystems(self):
    import sys
    print(" Calling _SetUpSystems from _integrator")
    print("I don't know what to do, I'm only the base class --> Terminate")
    sys.exit()


#------------------------------------------------------------------------------#


  def _SetUpPreconditioner(self):
    import sys
    print(" Calling _SetUpSystems from _integrator")
    print("I don't know what to do, I'm only the base class --> Terminate")
    sys.exit()


#------------------------------------------------------------------------------#


  def _CreateOutputFolder(self):
    from commix._tools.output import CreateNewFolder
    if (self._newFolder):
      output_dir = CreateNewFolder(self.output.foldername)
      print("Folder --", output_dir, "-- for saving results created.")
      self._newFolder = False
      self.output.foldername = output_dir
      
      if self.output.memoryWarning:
        videoMemoryWarning(mesh = self._geometry.GetMesh() \
          , simulationTime = self.parameters.T_end - self.parameters.T_start \
          , k_ = self.parameters.timeStepSize \
          , saveatevery = self.output.saveatevery)
      video_dir = CreateNewFolder(name = self.output.foldername + "/" \
        + self.output.foldernameVideo)
      print("Folder --", video_dir, "-- for saving video created.")
      self.output.foldernameVideo = video_dir
    else:
      print("Existing folder --", self.output.foldername \
        , "-- will be used for saving results.")


#------------------------------------------------------------------------------#


  def _ExistsMagnetization(self):
    import sys
    if (self._Mag == None):
      print("Error in _Integrator.ExistsMagnetization(self)")
      print("No magnetization specified. --> Terminate.")
      sys.exit()
  
  
#------------------------------------------------------------------------------#


  def Step(self):

    self._StepPrepare()
    self._StepSolve()
    self._StepUpdate()
    self._StepFollowUp()
  
  
#------------------------------------------------------------------------------#


  def _StepPrepare(self):
    from ngsolve import TaskManager

    self._StepPrepareBureaucracy()

#    with TaskManager(pajetrace=100*1000*1000):
    with TaskManager():
      self._StepPrepareWithTaskManager()

    self._StepPrepareWithoutTaskManager()



#------------------------------------------------------------------------------#


  def _StepPrepareBureaucracy(self):
    import time
    self._compTimer = time.time()
    
    self._ExistsMagnetization()

    self._ResetIterationCounter()


#------------------------------------------------------------------------------#


  def _StepPrepareWithTaskManager(self):
    import sys
    print("From _integrator._StepPrepareWithTaskManager():")
    print("I'm only the base class, I don't know what to do.")
    print("--> Terminate")
    sys.exit()


#------------------------------------------------------------------------------#


  def _StepPrepareWithoutTaskManager(self):
    import sys
    print("From _integrator._StepPrepareWithoutTaskManager():")
    print("I'm only the base class, I don't know what to do.")
    print("--> Terminate")
    sys.exit()


#------------------------------------------------------------------------------#


  def _AssembleSystem(self):
    self._AssembleLHS()
    self._AssembleRHS()


#------------------------------------------------------------------------------#


  def _ComputeMaxwell(self):

    self._maxwellTimer.Start()
    self._Mag.ComputeMaxwell(self._solvetol)
    self._maxwellTimer.Stop()


#------------------------------------------------------------------------------#


  def _StepSolve(self):
    import sys
    print("From _integrator._StepSolve():")
    print("I'm only the base class, I don't know how to Solve")
    print("--> Terminate")
    sys.exit()
    

#------------------------------------------------------------------------------#


  def _StepUpdate(self):
    import sys
    print("From _integrator._StepUpdate():")
    print("I'm only the base class, I don't know how to Solve")
    print("--> Terminate")
    sys.exit()
    

#------------------------------------------------------------------------------#


  def _StepFollowUp(self):
    import time
    self._compTimeNow = time.time()-self._compTimer
    self._compTimeTotal += self._compTimeNow

    self._Mag.PrintAverages(self.Time())
    
    if (self.output.draw):
      self._Mag.Draw()
    
    self._counter += 1
    self._t += self._k
    self._time_cf.Set( self._t ) 

    self.Write()
    


#------------------------------------------------------------------------------#

  
  def Integrate(self, duration=None, relax=False):

    tmp_H_ext = self.parameters.H_ext
    tmp_maxwellCoupling = self.parameters.maxwellCoupling
    tmp_spintronicsCoupling = self.parameters.spintronicsCoupling
    tmp_T_end = self.parameters.T_end

    if relax:
      self.parameters.H_ext = None
      self.SetParameter_maxwellCoupling("strayField")
      self.parameters.spintronicsCoupling = "none"

    if duration != None:
      if self.parameters.T_start == None:
        self.parameters.T_start = 0.0
      elif self._counter != 0:
        self.parameters.T_start = self.Time()
      self.parameters.T_end = self.parameters.T_start + duration
          
    self.GetReady()
    self.Do()

    self.parameters.H_ext = tmp_H_ext
    self.parameters.maxwellCoupling = tmp_maxwellCoupling
    self.parameters.spintronicsCoupling = tmp_spintronicsCoupling

    try:
      self.parameters.T_end = max(self.Time(), tmp_T_end)
    except:
      self.parameters.T_end = self.Time()


#------------------------------------------------------------------------------#


  def Do(self):
    self.SaveParameters()
    # TODO also a Debug of SaveParameters, for the math constants

    self._startCounter = self._counter
    if self._counter == 0: self.Write()

    if (self.output.draw): self._Mag.Draw()
    
    for step in range(self._steps):
      self.Step()
    
    print("Results have been written to: /", self.output.foldername)
  
  
#------------------------------------------------------------------------------#


  def SaveMagnetization(self, filename=None):
    from commix._tools.output import LiveSaveValues
    
    if( filename == None):
      self._Mag.Save(self.output.foldernameVideo + "/" \
          + self.output.filenameMagnetization + "_" + str(self._counter))
    
      if (self._counter == 0):
        timeHeader = "colums are: t,".ljust(self.output.columnWidth) \
          + "step_nr".ljust(self.output.columnWidth)
      else:
        timeHeader = None
          
      LiveSaveValues(self.output.foldernameVideo + "/" \
        + self.output.filenameVideotimes, self.Time() \
        , [self._counter], nrwidth = self.output.columnWidth \
        , header = timeHeader)
    else:
      self._Mag.Save(self.output.foldername + "/" + filename)
  
  
#------------------------------------------------------------------------------#


  def SaveMaxwell(self):
    
    if(self.parameters.maxwellCoupling == "strayField"):
      self._Mag.SaveStrayField(self.output.foldernameVideo + "/" \
          + self.output.filenameStrayField + "_" + str(self._counter))
    elif(self.parameters.maxwellCoupling == "oerstedField"):
      self._Mag.SaveOerstedField(self.output.foldernameVideo + "/" \
          + self.output.filenameOerstedField + "_" + str(self._counter))
    elif(self.parameters.maxwellCoupling == "magnetostaticField"):
      self._Mag.SaveStrayField(self.output.foldernameVideo + "/" \
          + self.output.filenameStrayField + "_" + str(self._counter))
      self._Mag.SaveOerstedField(self.output.foldernameVideo + "/" \
          + self.output.filenameOerstedField + "_" + str(self._counter))
    else:
      pass
  
  
#------------------------------------------------------------------------------#


  def SaveMesh(self, filename=None):
    
    if( filename == None):
      filename = self.output.filenameMesh
    
    self._geometry.Save( self.output.foldername + "/" + filename )
  
  
#------------------------------------------------------------------------------#


  def PrintTime(self):
    print("t = %.3e" % self.Time(), "s")
  
  
#------------------------------------------------------------------------------#


  def Write(self):
  
    # TODO make these conditional on counter, i.e., not write at every timestep
    self._WriteAverages()
    self._WritePointwiseAverages()
    self._WriteExternalField() 
    self._WriteCurrent() 
    self._WriteNorms()
    self._WriteComputationalTime()
    self._WriteEnergies()
    self._WriteMagnetization()
    self._WriteMaxwell()
    self._WriteIterations()


#------------------------------------------------------------------------------#


  def _WriteAverages(self):
    from commix._tools.output import LiveSaveValues
    ## averages
    headerAvg = self.output.headerAverages if self._counter == 0 else None
    avg = self._Mag.Averages()
    LiveSaveValues(self.output.foldername+"/"+self.output.filenameAverages \
      , self.Time(), avg, nrwidth=self.output.columnWidth, header=headerAvg)
    

#------------------------------------------------------------------------------#


  def _WritePointwiseAverages(self):
    from commix._tools.output import LiveSaveValues
    ## pointwise averages
    headerPointwiseAvg \
      = self.output.headerPointwiseAverages if self._counter == 0 else None
    pw_avg = self._Mag.PointwiseAverages()
    LiveSaveValues(self.output.foldername + "/" \
      + self.output.filenamePointwiseAverages \
      , self.Time(), pw_avg, nrwidth=self.output.columnWidth \
      , header=headerPointwiseAvg)
    

#------------------------------------------------------------------------------#


  def _WriteExternalField(self):
    from commix._tools.output import LiveSaveValues
    ## external field
    headerExt = self.output.headerExtField if self._counter == 0 else None
    ext = self._EvaluateExternalFieldAtPoint()
    LiveSaveValues(self.output.foldername+"/"+self.output.filenameExtField \
      , self.Time(), ext, nrwidth=self.output.columnWidth, header=headerExt)
    

#------------------------------------------------------------------------------#


  def _WriteCurrent(self):
    from commix._tools.output import LiveSaveValues
    ## current
    headerCur = self.output.headerCurrent if self._counter == 0 else None
    cur = self._EvaluateCurrent()
    LiveSaveValues(self.output.foldername+"/"+self.output.filenameCurrent \
      , self.Time(), cur, nrwidth=self.output.columnWidth, header=headerCur)
    

#------------------------------------------------------------------------------#


  def _WriteNorms(self):
    from commix._tools.output import LiveSaveValues
    ## norms
    headerNrm = self.output.headerNorms if self._counter == 0 else None
    nrm = self._Mag.Norms()
    nrm[1] *= self._geometry.scaling \
      * self._geometry.scaling*self._geometry.scaling
    LiveSaveValues(self.output.foldername+"/"+self.output.filenameNorms \
      , self.Time(), nrm, nrwidth=self.output.columnWidth, header=headerNrm)
    

#------------------------------------------------------------------------------#


  def _WriteComputationalTime(self):
    from commix._tools.output import LiveSaveValues
    ## computational Times
    headerCompTimes = self.output.headerCompTimes if self._counter == 0 else None
    LiveSaveValues(self.output.foldername+"/"+self.output.filenameCompTimes \
      , self.Time(), [self._compTimeNow, self._compTimeTotal] \
      , nrwidth=self.output.columnWidth \
      , header=headerCompTimes)
    

#------------------------------------------------------------------------------#


  def _WriteEnergies(self):
    from commix._tools.output import LiveSaveValues

    ## energies
    headerEnergy = self.output.headerEnergies if self._counter == 0 else None
    
    E_exch = self.parameters.A * self._geometry.scaling \
      * self._Mag.EnergyExchange()
    
    if (self.parameters.K == 0):
      E_ani = 0.0
    else:
      E_ani = self._geometry.scaling \
        * self._geometry.scaling * self._geometry.scaling \
        * self.parameters.K \
        * self._Mag.EnergyUniaxialAnisotropy(self._easy_cf)
    
    if ( (self.parameters.D == 0) or (self.parameters.dmCoupling == "none") ):
      E_dm = 0.0
    elif( self.parameters.dmCoupling == "bulk" ):
      E_dm = self._geometry.scaling * self._geometry.scaling \
        * self.parameters.D \
        * self._Mag.EnergyBulkDzyaloshinskiiMoriya()
    elif( self.parameters.dmCoupling == "interfacial" ):
      E_dm = self._geometry.scaling * self._geometry.scaling \
        * self.parameters.D \
        * self._Mag.EnergyInterfacialDzyaloshinskiiMoriya()
    
    if (self.parameters.maxwellCoupling == "strayField" \
      or self.parameters.maxwellCoupling == "magnetostaticField" ):
      E_stray = -0.5 * self.parameters.mu0 \
        * self.parameters.Ms * self.parameters.Ms \
        * self._geometry.scaling * self._geometry.scaling \
        * self._geometry.scaling * self._Mag.EnergyStrayField()
    else:
      E_stray = 0.0
    
    if (self.parameters.maxwellCoupling == "oerstedField" \
      or self.parameters.maxwellCoupling == "magnetostaticField" ):
      E_oersted = -0.5 * self.parameters.mu0 \
        * self.parameters.Ms * self.parameters.Ms \
        * self._geometry.scaling * self._geometry.scaling \
        * self._geometry.scaling * self._Mag.EnergyOerstedField()
    else:
      E_oersted = 0.0
    
#    if (self.parameters.maxwellCoupling == "eddyCurrent"):
#      E_eddy = 0.5 * self.parameters.mu0 \
#        * self.parameters.Ms * self.parameters.Ms \
#        * self._geometry.scaling * self._geometry.scaling \
#        * self._geometry.scaling * self._Mag.EnergyEddyCurrent()
#    else:
#      E_eddy = 0.0
    
    E_appl = -1.0 * self.parameters.mu0 \
      * self.parameters.Ms * self.parameters.Ms \
      * self._geometry.scaling * self._geometry.scaling \
      * self._geometry.scaling * self._Mag.EnergyApplied(self._hext_cf)
      
    E_total = E_exch + E_ani + E_dm + E_stray + E_oersted + E_appl #TOECDO + E_eddy
    
#TOECDO    en = [E_exch, E_ani, E_dm, E_stray, E_eddy, E_appl, E_total]
    en = [E_exch, E_ani, E_dm, E_stray, E_oersted, E_appl, E_total]
    
    LiveSaveValues(self.output.foldername+"/"+self.output.filenameEnergies \
      , self.Time(), en, nrwidth=self.output.columnWidth \
      , header=headerEnergy)
  

#------------------------------------------------------------------------------#


  def _WriteMagnetization(self):
    if self._counter == 0:
      self.SaveMagnetization(self.output.filenameMagnetization + "_" + "initial")
      self.SaveMesh()
    
    if self.output.video == False: return

    if ((self._counter - self._startCounter) % self.output.saveatevery == 0):
      self.SaveMagnetization()
    
  
#------------------------------------------------------------------------------#


  def _WriteMaxwell(self):
    if self.output.videoMaxwell == False: return

    if self._counter == 0:
      self._ComputeMaxwell()
      self.SaveMaxwell()
    
    if ((self._counter - self._startCounter) % self.output.saveatevery == 0):
      self.SaveMaxwell()
    
  
#------------------------------------------------------------------------------#


  def _WriteIterations(self):
    from commix._tools.output import LiveSaveValues
    if not isinstance(self._callback, _details.IterationCounter): return

    ## iterations
    headerItc = self.output.headerIterationCount if self._counter == 0 else None
    itc = [self._callback.GetCount()]
    LiveSaveValues(self.output.foldername+"/"+self.output.filenameIterationCount \
      , self.Time(), itc, nrwidth=self.output.columnWidth, header=headerItc)
    

#------------------------------------------------------------------------------#


  def Time(self):
    return self._t / self.parameters.Ms / self.parameters.gamma0
  
  
#------------------------------------------------------------------------------#


  def SaveParameters(self):
    from commix._tools.output import SaveContent
    # TODO: sort the output stuff, probably brute force?
    # TODO: I want output: "This is the 1./2./3. simualtion stage...
    saveString = ""
    if (self._counter == 0):
      saveString += self.output.headerParameters
      saveString += "\n" + "\n"
    
    saveString += str(self)
    saveString += "\n" + "\n"
    saveString += "########################################"
    saveString += "########################################"
    saveString += "\n" + "\n"
    
    SaveContent(self.output.foldername + "/" \
      + self.output.filenameParameters, saveString)
  
  
#------------------------------------------------------------------------------#


  def _EvaluateExternalFieldAtPoint(self):
    import types
    from ngsolve import x, y, z
    if self.parameters.H_ext == None:
      return [0.0, 0.0, 0.0]

    # TODO also full .vtk or chosen point (mip) should be possible
    ext = [0, 0, 0]
    for j in range(3):
      if isinstance(self.parameters.H_ext[j], types.FunctionType):
        time = self._time_cf / self.parameters.Ms / self.parameters.gamma0
        cf = self.parameters.H_ext[j](time, x, y, z)
        pnt = self.output.GetExternalFieldEvaluationPoint()
        mip = self._geometry.GetMesh()( \
          pnt[0] / self._geometry.scaling, \
          pnt[1] / self._geometry.scaling, \
          pnt[2] / self._geometry.scaling)
        ext[j] = cf(mip)
      else:
        ext[j] = self.parameters.H_ext[j]
    
    return ext


#------------------------------------------------------------------------------#


  def _EvaluateCurrent(self):
    import types
    import sys
    from ngsolve import CoefficientFunction
    if self.parameters.spintronicsCoupling == "none":
      return [0.0, 0.0, 0.0, 0.0]

    cur = [0, 0, 0, 0]
    if self.parameters.spintronicsCoupling == "slonczewski":
      if isinstance(self.parameters.Je, types.FunctionType):
        time = self._time_cf / self.parameters.Ms / self.parameters.gamma0
        cf = CoefficientFunction(self.parameters.Je(time))
        pnt = self._geometry.GetMesh().ngmesh.Points()[1].p
        mip = self._geometry.GetMesh()(pnt[0], pnt[1], pnt[2])
        cur[0] = cf(mip)
      else:
        cur[0] = self.parameters.Je
    elif self.parameters.spintronicsCoupling == "zhangLi":
      for j in range(3):
        if isinstance(self.parameters.Je[j], types.FunctionType):
          time = self._time_cf / self.parameters.Ms / self.parameters.gamma0
          cf = CoefficientFunction(self.parameters.Je[j](time))
          pnt = self._geometry.GetMesh().ngmesh.Points()[1].p
          mip = self._geometry.GetMesh()(pnt[0], pnt[1], pnt[2])
          cur[j+1] = cf(mip)
        else:
          cur[j+1] = self.parameters.Je[j]
    else:
      print("From _EvaluateCurrent(): spintronicsCoupling invalid")
      print("--> Terminate")
      sys.exit()
    
    return cur


#------------------------------------------------------------------------------#


  def _ZeroCF(self):
    from . import _specialForms
    return _specialForms.Zero()


#------------------------------------------------------------------------------#


  def SetResultsFolder(self, foldername):
    # TODO catch if changed during computation
    self.output.foldername = foldername


#------------------------------------------------------------------------------#


  def SetParameter_A(self, value):
    self.parameters.A = value


#------------------------------------------------------------------------------#


  def SetParameter_Ms(self, value):
    self.parameters.Ms = value


#------------------------------------------------------------------------------#


  def SetParameter_K(self, value):
    self.parameters.K = value


#------------------------------------------------------------------------------#


  def SetParameter_D(self, value):
    self.parameters.D = value


#------------------------------------------------------------------------------#


  def SetParameter_alpha(self, value):
    self.parameters.alpha = value


#------------------------------------------------------------------------------#


  def SetParameter_T_start(self, value):
    self.parameters.T_start = value


#------------------------------------------------------------------------------#


  def SetParameter_T_end(self, value):
    self.parameters.T_end = value


#------------------------------------------------------------------------------#


  def SetParameter_TimeStepSize(self, value):
    self.parameters.timeStepSize = value


#------------------------------------------------------------------------------#


  def SetParameter_H_ext(self, value):
    self.parameters.H_ext = value


#------------------------------------------------------------------------------#


  def SetParameter_easy_axis(self, value):
    self.parameters.easy_axis = value


#------------------------------------------------------------------------------#


  def SetParameter_P(self, value):
    self.parameters.P = value


#------------------------------------------------------------------------------#


  def SetParameter_J(self, value):
    self.parameters.J = value


#------------------------------------------------------------------------------#


  def SetParameter_xi(self, value):
    self.parameters.xi = value


#------------------------------------------------------------------------------#


  def SetParameter_dmCoupling(self, value):
    self.parameters.dmCoupling = value


#------------------------------------------------------------------------------#


  def SetParameter_maxwellCoupling(self, value):
    if value != self.parameters.maxwellCoupling:
      self._maxwellCouplingChanged = True
    self.parameters.maxwellCoupling = value


#------------------------------------------------------------------------------#


  def SetParameter_spintronicsCoupling(self, value):
    self.parameters.spintronicsCoupling = value


#------------------------------------------------------------------------------#


  def SetParameter_Je(self, value):
    self.parameters.Je = value


#------------------------------------------------------------------------------#


  def SetParameter_d(self, value):
    self.parameters.d = value


#------------------------------------------------------------------------------#


  def SetParameter_p(self, value):
    self.parameters.p = value


#------------------------------------------------------------------------------#


  def SetPreconditioner(self, value):
    self.parameters.preconditionerType = value


#------------------------------------------------------------------------------#


  def SetGuess(self, value):
    self.parameters.guessType = value


#------------------------------------------------------------------------------#


  def CountIterations(self, value=True):
    if value:
      self._callback = self._itCounter
    else:
      self._callback = None


#------------------------------------------------------------------------------#


  def _ResetIterationCounter(self):
    if isinstance(self._itCounter, _details.IterationCounter):
      self._itCounter.Reset()


#------------------------------------------------------------------------------#


  def SetTOL(self, value):
      self.parameters.TOL = value


#------------------------------------------------------------------------------#


  def RecordExternalFieldAt(self, value):
    self.output.SetExternalFieldEvaluationPoint(value)


#------------------------------------------------------------------------------#

  
  def _SetNumThreads(self, numthreads):
    import os
    os.environ['OMP_NUM_THREADS'] = str(numthreads)
    os.environ['NGS_NUM_THREADS'] = str(numthreads)
    os.environ['BEMPP_NUM_THREADS'] = str(numthreads)

    from mpi4py import MPI # import here once before bempp to avoid Rank error
    from ngsolve import ngsglobals
    ngsglobals.numthreads = numthreads


#------------------------------------------------------------------------------#


  def _RecordMagnetization(self, record, maxwell, every):
    self.output.video = record
    self.output.videoMaxwell = maxwell
    self._saveMagEvery = every


#------------------------------------------------------------------------------#


  def _SetUpOutput(self):
    if(self._saveMagEvery == None): return
    if(self._saveMagEvery == 0): return

    self.output.saveatevery = \
      round(self._saveMagEvery / self.parameters.timeStepSize)


#------------------------------------------------------------------------------#


################################################################################
