### TPS2 ###
from . import _tpsx


################################################################################


class TPS2(_tpsx.TPSX):
  # description of TPS2


#------------------------------------------------------------------------------#

  
  def __init__(self, geometry, parameters = None, output = None, numthreads = 1):

    _tpsx.TPSX.__init__(self, geometry, parameters, output, numthreads)

    from scipy import log
    self.rho = lambda q : q*abs(log(q))
    self.MWeightCut = lambda q : 1.0 / (q*abs(log(q)))

    self._theta = 0.5

    self._maxiter = 20


#------------------------------------------------------------------------------#


  def GetReady(self):
    self._itertol = self.parameters.TOL
    self._solvetol = self.parameters.TOL / 100.0

    _tpsx.TPSX.GetReady(self)
    self._Mag.Project()


#------------------------------------------------------------------------------#


  def _CheckForBadParameterCombinations(self):
    if (self.parameters.theta != 0.5):
      raise ValueError("theta!=0.5 not available for", type(self))

    _tpsx.TPSX._CheckForBadParameterCombinations(self)


#------------------------------------------------------------------------------#


  def _SetRHS_3D(self):
    self._RHS_3D = self._f_fix.vec.FV().NumPy() + self._f_iter.vec.FV().NumPy()


#------------------------------------------------------------------------------#


  def _StepUpdate(self):
    self._Mag.m.vec.data += self._k * self._Mag.v.gf.vec
    self._Mag.Project()


#------------------------------------------------------------------------------#


  def _SetUpSystems(self):
    from commix._tools import ScipySparseFromNgMatrix
    from ngsolve import TaskManager, TET, IntegrationRule, SymbolicLFI, LinearForm, SymbolicBFI, BilinearForm

    #blf
    lhs_stat_cf = self._StiffTerm()
    lhs_instat_cf = self._MassTerm() + self._SkewTerm()
    
    self._a_stat = BilinearForm(self._X)
    self._a_stat += SymbolicBFI( lhs_stat_cf.Compile(realcompile=True, wait=True) )
    
    with TaskManager():
      self._a_stat.Assemble()
    self._A_stat = ScipySparseFromNgMatrix(self._a_stat)
    
    ir3 = IntegrationRule(TET, order=3)
    
    self._a_instat = BilinearForm(self._X)
    self._a_instat += SymbolicBFI( lhs_instat_cf.Compile(realcompile=True, wait=True), intrule=ir3 )
    
    with TaskManager():
      self._a_instat.Assemble()
    self._A_instat = ScipySparseFromNgMatrix(self._a_instat) 

    #lf
    rhs_fix_cf = self._ExchangeTerm()
    rhs_fix_cf += self._LinearTerms_pi()
    rhs_fix_cf += self._NonLinearTerms_Pi()
    rhs_fix_cf += self._AppliedTerm()

    self._f_fix = LinearForm(self._X)
    self._f_fix += SymbolicLFI(rhs_fix_cf.Compile(realcompile=True, wait=True))

    rhs_iter_cf = self._DerivLinearTerms_pi()
    rhs_iter_cf += self._DerivNonLinearTerms_Pi()
    
    self._f_iter = LinearForm(self._X)
    self._f_iter += SymbolicLFI(rhs_iter_cf.Compile(realcompile=True, wait=True))


#------------------------------------------------------------------------------#


  def _ComputeWeight(self):

    tmp_cf = self._WeightExchange()

    if (self.parameters.H_ext != None and any(self.parameters.H_ext)):
      tmp_cf += self._WeightApplied()

    if (self._C_ani):
      tmp_cf += self._WeightUniaxialAnisotropy()

    if (self.parameters.dmCoupling != "none"):
      tmp_cf += self._WeightDzyaloshinskiiMoriya()

    if(self.parameters.maxwellCoupling != "none"):
      tmp_cf += self._WeightMaxwell()

    if(self.parameters.spintronicsCoupling != "none"):
      tmp_cf += self._WeightSpintronics()

    return self._WeightCut(tmp_cf)


#------------------------------------------------------------------------------#


  def _Compute_v(self):
    import sys
    from scipy import sqrt
    import scipy.sparse.linalg
    from ngsolve import Integrate, GridFunction
    import numpy as np
    if( self.parameters.maxwellCoupling == "magnetostaticField" ):
      self._Mag.v.hs.SetZero()
      self._Mag.v.hc.SetZero()
    elif( self.parameters.maxwellCoupling == "strayField" ):
      self._Mag.v.hs.SetZero()
    elif( self.parameters.maxwellCoupling == "oerstedField" ):
      self._Mag.v.hc.SetZero()
    elif( self.parameters.maxwellCoupling == "eddyCurrent" ):
      pass #TOECDO
#      self.H.SetEtaToH()
    elif (self.parameters.maxwellCoupling != "none"):
      print("Error in ._Mag.Compute_v_implicit():")
      print("Maxwell options not >strayField<, >oerstedField<, >eddyCurrent< \
        or >none< are not implemented yet")
      sys.exit()
    
    # assemble blf and lf
    self._Assemble_A_instat()
    
    A = self._A_stat + self._A_instat
    
    itnr = 0 
    oldIterate = np.array( self._Mag.v.gf.vec.FV() )
    
    diff_gf = GridFunction(self._X)
    diff_cf = diff_gf
    
    self._Assemble_f_fix()

    for itnr in range(self._maxiter):
      self._Assemble_f_iter()
      f = self._f_fix.vec.FV().NumPy() + self._f_iter.vec.FV().NumPy()

      self._Compute_Q()
      LHS = self._Q.dot( A.dot( self._Q.transpose() ) ) 
      RHS = self._Q.dot( f )
    
      w, succ = scipy.sparse.linalg.gmres(LHS, RHS , x0=None \
        , tol=self._solvetol, maxiter=4000, M=None)
      self._Mag.v.gf.vec.FV().NumPy()[:] = self._Q.transpose().dot(w)
    
      diff_gf.vec.FV().NumPy()[:] = self._Mag.v.gf.vec.FV().NumPy()-oldIterate
      itError = sqrt( Integrate(diff_cf*diff_cf, self._GetMesh()) )
    
      if ( self.parameters.maxwellCoupling == "eddyCurrent" ):
        pass # TOECDO
#        self.H.SolveForEta(self.m, self.v.gf)
#        itError += self.H.IterationError()
    
      # TOECDO consider different error for Eddy Currents?
      if (itError <= self._itertol):
        break
    
      print("itnr = ", itnr, "; Err = ", itError)
    
      oldIterate[:] = self._Mag.v.gf.vec.FV().NumPy()
      self._Mag.v.ComputeMaxwell(self._solvetol)

    print("itnr = ", itnr, "; Err = ", itError)


#------------------------------------------------------------------------------#


  def _Assemble_f_fix(self):
    from ngsolve import TaskManager
    with TaskManager():
      self._f_fix.Assemble()


#------------------------------------------------------------------------------#


  def _Assemble_f_iter(self):
    from ngsolve import TaskManager
    with TaskManager():
      self._f_iter.Assemble()


#------------------------------------------------------------------------------#


  def _MassTerm(self):
    return self._ComputeWeight() * self._Mass()


#------------------------------------------------------------------------------#


  def _StiffTerm(self):
    return self._C_ex * self._theta * self._k * (1.0 + self.rho(self._k)) \
      * self._Stiff()


#------------------------------------------------------------------------------#


  def _Applied(self):
    from . import _linearForms
    if self.parameters.H_ext == None:
      return self._ZeroCF()
    else:
      return _linearForms.Applied(self._phi, self._hext_impl_cf)


#------------------------------------------------------------------------------#

  
  def _DerivLinearTerms_pi(self):
    return self._theta * self._k * \
      ( self._DerivUniaxialAnisotropyTerm() \
      + self._DerivMaxwellTerm() \
      + self._DerivDzyaloshinskiiMoriyaTerm() )


#------------------------------------------------------------------------------#


  def _DerivNonLinearTerms_Pi(self):
    return self._theta * self._k * self._DerivSpintronicsTerm()


#------------------------------------------------------------------------------#


  def _DerivUniaxialAnisotropyTerm(self):
    return -self._C_ani * self._DerivUniaxialAnisotropy()


#------------------------------------------------------------------------------#


  def _DerivMaxwellTerm(self):
    return self._DerivMaxwell()


#------------------------------------------------------------------------------#


  def _DerivDzyaloshinskiiMoriyaTerm(self):
    return -0.5 * self._C_dm * self._DerivDzyaloshinskiiMoriya()


#------------------------------------------------------------------------------#


  def _DerivSpintronicsTerm(self):
    return self._C_st \
      * self._DerivSpintronics(self._Mag.m, self._Mag.v.gf \
      , self._Mag.Dm, self._Mag.v.Dgf)


#------------------------------------------------------------------------------#


  def _DerivUniaxialAnisotropy(self):
    return self._UniaxialAnisotropy(self._Mag.v.gf)


#------------------------------------------------------------------------------#


  def _DerivMaxwell(self):
    # TOECDO Achtung funktioniert so nur fuer Streufeld
    return self._Maxwell(self._Mag.v)


#------------------------------------------------------------------------------#


  def _DerivDzyaloshinskiiMoriya(self):
    return self._DzyaloshinskiiMoriya(self._Mag.v.gf, self._Mag.v.Dgf)


#------------------------------------------------------------------------------#


  def _DerivSpintronics(self, gf1, gf2, Dgf1, Dgf2):
    import sys
    if(self.parameters.spintronicsCoupling == "none"):
      return self._ZeroCF()

    if(self.parameters.spintronicsCoupling == "zhangLi"):
      return self._DerivZhangLi(gf1, gf2, Dgf1, Dgf2)
    elif(self.parameters.spintronicsCoupling == "slonczewski"):
      return self._DerivSlonczewski(gf1, gf2)
    else: 
      print("This option is not implemented.")
      sys.exit()
    

#------------------------------------------------------------------------------#


  def _DerivZhangLi(self, gf1, gf2, Dgf1, Dgf2):
    from . import _linearForms
    return _linearForms.ZhangLiFirst(self._phi, self._j_cf, Dgf1, gf2) \
      + _linearForms.ZhangLiFirst(self._phi, self._j_cf, Dgf2, gf1) \
      + _linearForms.ZhangLiFirst(self._phi, self._jprime_cf, Dgf1, gf1) \
      + self.parameters.xi \
      * _linearForms.ZhangLiSecond(self._phi, self._j_cf, Dgf2) \
      + self.parameters.xi \
      * _linearForms.ZhangLiSecond(self._phi, self._jprime_cf, Dgf1)


#------------------------------------------------------------------------------#


  def _DerivSlonczewski(self, gf1, gf2):
    from . import _linearForms
    return _linearForms.SlonczewskiSecond(self._phi, self._j_cf \
      , self.parameters.P, self._p_cf, gf1, gf2) \
      + _linearForms.SlonczewskiFirst(self._phi, self._j_cf, self.parameters.P \
      , self._p_cf, gf1, gf2) \
      + _linearForms.SlonczewskiFirst(self._phi, self._jprime_cf \
      , self.parameters.P, self._p_cf, gf1, gf1)


#------------------------------------------------------------------------------#


  @staticmethod
  def _IsMagProjected():
    return True


#------------------------------------------------------------------------------#


  def _WeightExchange(self):
    from commix._tools.algebra import FrobeniusInnerProd
    return -self._C_ex * FrobeniusInnerProd(self._Mag.Dm, self._Mag.Dm)
    

#------------------------------------------------------------------------------#


  def _WeightApplied(self):
    from commix._tools.algebra import InnerProd
    return InnerProd(self._hext_cf, self._Mag.m)
    

#------------------------------------------------------------------------------#


  def _WeightUniaxialAnisotropy(self):
    from commix._tools.algebra import InnerProd
    return -self._C_ani \
      * InnerProd(self._Mag.phiUniaxialAnisotropy(self._easy_cf), self._Mag.cf)
    

#------------------------------------------------------------------------------#


  def _WeightDzyaloshinskiiMoriya(self):
    import sys
    from commix._tools.algebra import Curl
    from commix._tools.algebra import InnerProd
    from . import _linearForms
    if(self.parameters.dmCoupling == "bulk"):
      return -self._C_dm \
        * InnerProd(Curl(self._Mag.Dm), self._Mag.m)
    elif(self.parameters.dmCoupling == "interfacial"):
      from commix.integrators._linearForms._opInterfacialDMI \
        import _OpInterfacialDMI
      return -self._C_dm \
        * InnerProd(self._Mag.m \
        , _OpInterfacialDMI(self._Mag.Dm, self.parameters.dmi_axis))
    elif (self.parameters.dmCoupling != "none"):
      print("Error in TPS2._ComputeWeight():")
      print("Dzyaloshinskii-Moriya options not >bulk<, >interfacial< \
        or >none< are not implemented yet")
      sys.exit()
    

#------------------------------------------------------------------------------#


  def _WeightMaxwell(self):
    import sys
    from commix._tools.algebra import InnerProd
    if (self.parameters.maxwellCoupling=="magnetostaticField"):
      return InnerProd(self._Mag.hs.cf, self._Mag.m) \
        + InnerProd(self._Mag.hc.cf, self._Mag.m)
    if (self.parameters.maxwellCoupling=="strayField"):
      return InnerProd(self._Mag.hs.cf, self._Mag.m)
    if (self.parameters.maxwellCoupling=="oerstedField"):
      return InnerProd(self._Mag.hc.cf, self._Mag.m)
    elif (self.parameters.maxwellCoupling != "none"):
      print("Error in TPS2.ComputeWeight():")
      print("Maxwell options not >strayField<, >oerstedField<, \
        or >none< are not implemented yet")
      sys.exit()
    

#------------------------------------------------------------------------------#


  def _WeightSpintronics(self):
    import sys
    if(self.parameters.spintronicsCoupling == "zhangLi"):
      return self._WeightZhangLi()
    elif(self.parameters.spintronicsCoupling == "slonczewski"):
      return self._WeightSlonczewski()
    elif(self.parameters.spintronicsCoupling != "none"):
      print("Error in TPS2._ComputeWeight():")
      print("Spintronics options not >zhangLi<, >slonczewski< \
        or >none< are not implemented yet")
      sys.exit()


#------------------------------------------------------------------------------#


  def _WeightZhangLi(self):
    from commix._tools.algebra import CrossProd
    from ngsolve import CoefficientFunction
    return self._C_st * CoefficientFunction( CrossProd( self._Mag.m \
      , self._Mag.opZhangLi(self._j_cf) ) ) * self._Mag.cf \
      + self._C_st * self.parameters.xi \
      * CoefficientFunction(self._Mag.opZhangLi(self._j_cf) ) \
      * self._Mag.cf
    

#------------------------------------------------------------------------------#


  def _WeightSlonczewski(self):
    from commix._tools.algebra import CrossProd
    from ngsolve import CoefficientFunction
    from . import _linearForms
    from commix.integrators._linearForms._gSlonczewski import _G
    G = _G(self._j_cf, self.parameters.P, self._p_cf, self._Mag.m)

    return self._C_st * G * CoefficientFunction( CrossProd( self._Mag.m \
      , self._p_cf ) ) * self._Mag.cf


#------------------------------------------------------------------------------#


  def _WeightCut(self, cf):
    from ngsolve import IfPos
    Min = lambda a, b : IfPos(a-b, b, a)

    case_pos = self.parameters.alpha + self._k*self._theta \
      * Min(cf, self.MWeightCut(self._k))
    case_neg = self.parameters.alpha \
      / (1.0 + self._k*self._theta/self.parameters.alpha \
      * Min(-cf, self.MWeightCut(self._k)))

    return IfPos(cf, case_pos, case_neg)


#------------------------------------------------------------------------------#


################################################################################
